/* wayV - NGUI (Next Generation User Interfaces)
 *
 * Copyright (C) 2000-2001 Mike Bennett (smoog@stressbunny.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Purpose:
 *	Simple image grabber and streamer using the Video4Linux API
 *
 * NOTE:
 *	This is very early code and I'd be surprised if it works
 *	for anyone but me (and then only barely)
*/
#include<stdlib.h>
#include<unistd.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<sys/ioctl.h>
#include<stdio.h>
#include<X11/keysym.h>
#include<X11/Xlib.h>
#include<X11/Xutil.h>

#include <linux/videodev.h>

#define VIDEODEV	"/dev/video0"
#define	NOISE		30
#define MAX_POINTS	1000

int main(int argc, char *argv[]) {
	struct video_capability cap;
	struct video_picture vpic;
	struct video_window win;
	int fd, size, screen, i, a, b, avg, x, y;
	unsigned char *image_data, *image_base, *image_diff, *image_noise;
	Display *display;
	GC gcontext;
	Window window;
	XImage *ximage, *xbase, *xdiff, *xnoise;
	int x_array[MAX_POINTS], y_array[MAX_POINTS], no_array;

	/* Try and access the webcam */
	if((fd = open(VIDEODEV, O_RDONLY)) < 0) {
		perror(VIDEODEV);
		exit(1);
	}

	/* Try and open the X display */
	display = XOpenDisplay(XDisplayName(NULL));

	if(!display) {
		printf("Cannot open the display\n");
		exit(-1);
	}

	/* Figure out the video device capabilities */
	ioctl(fd, VIDIOCGCAP, &cap);

	/* Figure out the image properties and make sure using the right palette */
	ioctl(fd, VIDIOCGPICT, &vpic);
	vpic.palette = VIDEO_PALETTE_RGB24;
	ioctl(fd, VIDIOCSPICT, &vpic);

	/* Figure out the input window setup */
	ioctl(fd, VIDIOCGWIN, &win);

	printf("Width = %d, height = %d, depth = %d\n", win.width, win.height,
		vpic.depth);

	size = win.width * win.height * (vpic.depth / 8);
	image_data = (unsigned char *)malloc(size);
	image_base = (unsigned char *)malloc(size);
	image_diff = (unsigned char *)malloc(size);
	image_noise = (unsigned char *)malloc(size);

	/* Setup the X display */
	screen = DefaultScreen(display);

	printf("Depth: %d\n", DefaultDepth(display, screen));

	window = XCreateSimpleWindow(display, RootWindow(display, screen),
			10, 10, (win.width * 2) + 1, (win.height * 2) + 1, 10,
			WhitePixel(display, screen), BlackPixel(display, screen));

	gcontext = XCreateGC(display, window, 0, NULL);

	XMapRaised(display, window);

	XFlush(display);

	ximage = (XImage *)XCreateImage(display, DefaultVisual(display, screen), 
			DefaultDepth(display, screen), ZPixmap, 0, image_data,
			win.width, win.height, 8, 0);

	xbase = (XImage *)XCreateImage(display, DefaultVisual(display, screen), 
			DefaultDepth(display, screen), ZPixmap, 0, image_base,
			win.width, win.height, 8, 0);

	xdiff = (XImage *)XCreateImage(display, DefaultVisual(display, screen), 
			DefaultDepth(display, screen), ZPixmap, 0, image_diff,
			win.width, win.height, 8, 0);

	xnoise = (XImage *)XCreateImage(display, DefaultVisual(display, screen), 
			DefaultDepth(display, screen), ZPixmap, 0, image_noise,
			win.width, win.height, 8, 0);

	/* Let the camera adjust for brightness, etc by capturing 25 dummy frames */
	read(fd, xbase->data, size);
	XPutImage(display, window, gcontext, xbase, 0, 0, 0, 0,
		win.width, win.height);
	read(fd, xbase->data, size);
	XFlush(display);

	for(b = 0; b < MAX_POINTS; b++) {
		x_array[b] = -1;
		y_array[b] = -1;
	}

	while(1) {
		read(fd, ximage->data, size);

		XPutImage(display, window, gcontext, ximage, 0, 0, 0,
			0, win.width, win.height);

		XPutImage(display, window, gcontext, xbase, 0, 0, win.width + 1,
			0, win.width, win.height);

		/* Create a movement map, find most significant difference regions */
		for(no_array = 0, a = 0, i = 0; i < size; i++) {

			for(avg = 0, b = i + 3; i < b; i++) {
				xnoise->data[i] = 255;
				xdiff->data[i] = ximage->data[i] - xbase->data[i];
				xbase->data[i] = ximage->data[i];

				if(xdiff->data[i] < 0)
					xdiff->data[i] = 0;

				avg += xdiff->data[i];
			}

			avg /= 3;

			if(avg > 0) {
				a++;

				if(a > NOISE) {
					y = (i - (3 * NOISE)) / (win.width * (vpic.depth / 8));
					x = ((i - (3 * NOISE)) - (y * win.width * (vpic.depth / 8))) / (vpic.depth / 8);

					if(no_array < MAX_POINTS) {
						x_array[no_array] = x;
						y_array[no_array] = y;
						a = 0;
						no_array++;
					}
				}
			} else {
				a = 0;
			}
		}

		a = win.width + 1;
		b = win.height + 1;

		for(i = 0; i < MAX_POINTS && x_array[i] != -1; i++) {

			if(a >= x_array[i] && b >= y_array[i]) {
				a = x_array[i];
				b = y_array[i];

				printf("a = %d, b = %d\n", a, b);
			}
		}

		XClearArea(display, window, a + win.width + 1, b + win.height + 1,
				10, 10, False);

		for(b = 0; b < MAX_POINTS; b++) {
			x_array[b] = -1;
			y_array[b] = -1;
		}

		XPutImage(display, window, gcontext, xdiff, 0, 0, 0, win.height + 1,
			win.width, win.height);

		XPutImage(display, window, gcontext, xnoise, 0, 0, win.width + 1,
			win.height + 1, win.width, win.height);

		XFlush(display);
	}
	
	close(fd);
	
	return 0;
}
